// vim:syntax=ragel
/* 
 * Pike CMOD parser for JSON.
 * Copyright (C) 2007 Arne Goedeke
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdio.h>
#include "global.h"
#include "interpret.h"
#include "stralloc.h"
#include "mapping.h"
#include "svalue.h"
#include "operators.h"
#include "array.h"
#include "builtin_functions.h"
#include "module.h"
#include "gc.h"

#include "json.h"

p_wchar2 *_parse_JSON(p_wchar2* p, p_wchar2* pe, struct parser_state *state); 

#include "json_string.c"
#include "json_number.c"
#include "json_array.c"
#include "json_mapping.c"
#ifdef JSON_CUSTOM_TYPES
# include "json_custom.c"
#endif

%%{
    machine JSON;
    alphtype int;
    write data;

    number_start = [\-+.] | digit;
    array_start = '[';
    mapping_start = '{';
    string_start = '"';
    myspace = [ \n\r\t];
    json_true = 'true';
    json_false = 'false';
    json_null = 'null';
    value_start = number_start | array_start | mapping_start | string_start | 'n' | 't' | 'f';

    action parse_string { PARSE(string, fpc); fexec i; }
    action parse_number { PARSE(number, fpc); fexec i; }
    action parse_mapping { PARSE(mapping, fpc); fexec i; }
    action parse_array { PARSE(array, fpc); fexec i; }
#action parse_custom { PARSE(custom, fpc); fexec i; }

    action push_true { PUSH_SPECIAL("true"); }
    action push_false { PUSH_SPECIAL("false"); }
    action push_null { PUSH_SPECIAL("null"); }

    main := myspace* . (number_start >parse_number |
			string_start >parse_string |
			mapping_start >parse_mapping |
			array_start >parse_array |
			json_true @push_true |
			json_false @push_false |
			json_null @push_null
#			| (any - (value_start | myspace)) >parse_custom
			) . myspace* %*{ if (state->level != 0) fbreak; };
}%%

p_wchar2 *_parse_JSON(p_wchar2 *p, p_wchar2 *pe, struct parser_state *state) {
    p_wchar2 *i = p;
    int cs;
    int c = 0;

    %% write init;
    %% write exec;

    if (cs >= JSON_first_final) {
	return p;
    }

    if (!state->validate && c > 0) pop_n_elems(c);

#ifdef JUMP
    if (!state->validate) {
	Pike_error("Error parsing JSON at '%c'\n", (char)*p);
    }
#endif

    push_int((int)p);
    return NULL;
}

/*! @module Public
 */

/*! @module Parser
 */

/*! @module JSON2
 *! This module contains a parser to parse JSON into native pike types. The parser has been written
 *! in c using ragel (@[http://www.cs.queensu.ca/~thurston/ragel/]).
 *! The parser is supposed to handle Unicode strings (internally 8, 16 and 32 bit wide strings that is).
 *! 
 *! Have a look at @[http://www.json.org] or @[http://www.ietf.org/rfc/rfc4627.txt?number=4627] for
 *! information about what JSON is.
 */

/*! @decl int validate(string s)
 *!
 *! Takes a string and checks if it is valid JSON.
 *! 
 *! @returns
 *! 	In case the string contains valid JSON @expr{-1@} is returned. It is then guarenteed to be parsed
 *! 	without errors by @[parse()].
 *! 	In case the string is not valid JSON, the integer position inside the string
 *! 	where the error occures is returned.
 */
PIKEFUN int validate(string data) {
    p_wchar2 *ret;
    struct parser_state state;

    JSON_CONVERT(data, ret);

    state.validate = 1;
    state.level = 0;
#ifdef JSON_CUSTOM_TYPES
    state.custom_chars = NULL;
#endif
    state.start = ret;
    state.stop = ret + data->len;
    state.data = data;

    pop_stack();
    if (_parse_JSON(ret, ret + data->len, &state) == NULL) {
	push_int((int)ret);
	f_minus(2);
	push_int(4);
	o_divide();
    } else {
	push_int(-1);
    }

    free(ret);
    return;
}

/*! @decl array|mapping|string|float|int parse(string s)
 *!
 *! Parses a JSON-formatted string and returns the corresponding pike data type.
 *! 
 *! @throws
 *! 	Throws an exception in case the data contained in @expr{s@} is not valid
 *! 	JSON.
 */
PIKEFUN mixed parse(string data) {
    p_wchar2 *ret, *stop;
    struct parser_state state;
    // we wont be building more than one string at once.
    JSON_CONVERT(data, ret);

    state.validate = 0;
    state.level = 0;
#ifdef JSON_CUSTOM_TYPES
    state.custom_chars = NULL;
#endif
    state.start = ret;
    state.stop = ret + data->len;
    state.data = data;

    stop = _parse_JSON(ret, ret + data->len, &state);

    if (stop != NULL) {
	free(ret);
	stack_pop_keep_top();

	return;
    }

    push_int((int)ret);
    // calculate offset in string where parsing failed
    f_minus(2);
    push_int(4);
    // divide by 4 to get offset in chars
    o_divide();
    push_int(((struct svalue*)(Pike_sp - 1))->u.integer);
    push_int(10);
    // we want 10 chars..
    f_add(2);
    f_index(3);
    free(ret);
    if (((struct svalue*)(Pike_sp - 1))->type != PIKE_T_STRING) {
	Pike_error("Parsing JSON failed and I dont know where.\n");
    } else {
	push_text("Parsing JSON failed at '%s'.\n");
	stack_swap();
	f_sprintf(2);
	f_aggregate(1);
	f_throw(1);
    }
}

/*! @endmodule
 */
/*! @endmodule
 */
/*! @endmodule
 */
